from copy import deepcopy

import numpy as np
import torch

from ..data import FlatTransition
from . import OffPolicyAlgorithm
from .goal_sampling_strategy import GoalSamplingStrategy

class HAC(OffPolicyAlgorithm):
    """Hierarchical Actor-Critic."""

    supported_goal_sampling_strategies = {"future", "episode", "final"}

    def __init__(self, name, model, child_failure_penalty, check_achievement, 
            flat_algo_kwargs=None, flat_algo_name="SAC", goal_sampling_strategy="future", 
            buffer_size=100000, testing_buffer_size=50000, batch_size=128, n_hindsight_goals=4, 
            testing_fraction=0.3, fully_random_fraction=0.1, bootstrap_testing_transitions=True, 
            use_normal_trans_for_testing=False, use_testing_transitions=True, 
            learn_from_deterministic_episodes=True, log_q_values=True, learning_starts=0, 
            bootstrap_end_of_episode=True, grad_steps_per_env_step=1):
            """
            Args:
                check_achievement (callable): Maps achieved_goal, 
                    desired_goal and parent_info to boolean achieved and float
                    reward indicating whether the achieved_goal satisfies the 
                    desired goal and what reward this implies.
                """

            super(HAC, self).__init__(name, flat_algo_name, model, fully_random_fraction, 
                    flat_algo_kwargs, buffer_size, batch_size, learning_starts,
                    grad_steps_per_env_step)


            assert goal_sampling_strategy in self.supported_goal_sampling_strategies\
                    or isinstance(goal_sampling_strategy, GoalSamplingStrategy), \
                    "Goal sampling strategy {} not supported.".format(goal_sampling_strategy)

            self._child_failure_penalty = child_failure_penalty
            self._goal_sampling_strategy = goal_sampling_strategy
            self._check_achievement = check_achievement
            self._n_hindsight_goals = n_hindsight_goals
            self._testing_fraction = testing_fraction
            self._bootstrap_testing_transitions = bootstrap_testing_transitions

            self._use_normal_trans_for_testing = use_normal_trans_for_testing
            self._use_testing_transitions = use_testing_transitions
            self._learn_from_deterministic_episodes = learn_from_deterministic_episodes
            self._log_q_values = log_q_values
            self._bootstrap_end_of_episode = bootstrap_end_of_episode

            self._verbose = False

    def _sample_achieved_goals(self, current_index):
        goals = []
        if self._n_hindsight_goals > 0:
            n_transitions = len(self._episode_transitions)
            if self._goal_sampling_strategy == "future":
                n_goals = min(self._n_hindsight_goals, 
                        n_transitions - current_index - 1)
                indices = np.random.randint(low = current_index + 1, high = n_transitions, 
                        size = n_goals)
            elif self._goal_sampling_strategy == "episode":
                n_goals = min(self._n_hindsight_goals, n_transitions)
                indices = np.random.randint(low = 0, high = n_transitions, size = n_goals)
            elif self._goal_sampling_strategy == "final":
                # only generate hindsight goal from final state if environment is done
                if self._episode_transitions[-1].env_done:
                    indices = [-1]
                else:
                    indices = []
            elif isinstance(self._goal_sampling_strategy, GoalSamplingStrategy):
                indices = self._goal_sampling_strategy(self._episode_transitions, 
                        self._n_hindsight_goals)

            for i in indices:
                goals.append(self._episode_transitions[i].subtask_tr.info["achieved_generalized_goal"])

        return goals

    def _add_experience_to_flat_algo(self, parent_info, deterministic_episode, node_is_sink, sess_info):
        if self._learn_from_deterministic_episodes or not deterministic_episode:
            self._ep_return = 0
            # add transitions of this episode to replay buffer of flat RL algorithm
            # (by applying hindsight goal and action manipulations)
            for trans_index, tr in enumerate(self._episode_transitions):
                has_achieved_now = tr.subtask_tr.info.get("has_achieved", False)

                if self._bootstrap_end_of_episode:
                    done = has_achieved_now
                else:
                    done = has_achieved_now or tr.env_info.done

                # unaltered flat transition
                f_trans_0 = FlatTransition(
                        obs = tr.subtask_tr.obs, 
                        action = tr.subtask_tr.action,
                        reward = tr.subtask_tr.reward,
                        new_obs = tr.subtask_tr.new_obs,
                        done = done)

                # if the node is a sink, do not attempt to manipulate action
                # in hindsight and add original transition to replay buffer
                if node_is_sink:
                    testing_transition = False
                    f_trans_base = f_trans_0
                    self._add_to_flat_replay_buffer(f_trans_0)
                    self._ep_return += f_trans_0.reward

                # if the node is not a sink consider testing transitions
                # and hindsight action transitions
                else:
                    # did the child node use a deterministic version of its policy?
                    # (testing transition)
                    testing_transition = tr.algo_info["child_be_deterministic"]

                    # boolean indicating whether child achieved subgoal
                    did_child_achieve_subgoal = tr.child_feedback["has_achieved"]

                    # add testing transitions (if enabled)
                    # Optionally, also transitions generated by stochastic 
                    # lower level are used as testing transitions.
                    # Only add transition with penalty if child node failed 
                    # to achieve its subgoal, otherwise do not add any transition.
                    if self._use_testing_transitions \
                       and (testing_transition or self._use_normal_trans_for_testing) \
                       and not did_child_achieve_subgoal:
                        f_trans_testing = deepcopy(f_trans_0)
                        f_trans_testing.reward = self._child_failure_penalty
                        if not self._bootstrap_testing_transitions:
                            # Have to set done to True in these transitions 
                            # (according to HAC paper and accompanying code).
                            f_trans_testing.done = True
                        if self._verbose:
                            print("testing transition in {}\n".format(self.name) 
                                    + str(f_trans_testing))
                        self._add_to_flat_replay_buffer(f_trans_testing)
                        self._ep_return += f_trans_testing.reward

                    # hindsight action transition
                    # If child node achieved desired goal use original action.
                    # If not, use subgoal the child achieved as action.
                    f_trans_hindsight_action = deepcopy(f_trans_0)
                    if not testing_transition or did_child_achieve_subgoal:
                        self._ep_return += f_trans_hindsight_action.reward
                    if not did_child_achieve_subgoal:
                        f_trans_hindsight_action.action = tr.child_feedback["achieved_generalized_goal"]
                        # TODO: In principle, reward could depend on action, so have to recompute
                        if self._verbose:
                            print("hindsight action transition in {}\n".format(self.name) 
                                    + str(f_trans_hindsight_action))
                    self._add_to_flat_replay_buffer(f_trans_hindsight_action)
                    f_trans_base = f_trans_hindsight_action

                # hindsight goal transitions (based on hindsight action 
                # transition or original transition in case of a sink node)
                achieved_goals = self._sample_achieved_goals(trans_index)
                for hindsight_goal in achieved_goals:
                    f_trans_hindsight_goal = deepcopy(f_trans_base)
                    f_trans_hindsight_goal.obs = {
                        "partial_observation": tr.subtask_tr.obs["partial_observation"], 
                        "desired_goal": hindsight_goal
                    }
                    f_trans_hindsight_goal.new_obs = {
                        "partial_observation": tr.subtask_tr.new_obs["partial_observation"], 
                        "desired_goal": hindsight_goal
                    }
                    achieved, f_reward = self._check_achievement(
                            achieved_goal = tr.subtask_tr.info["achieved_generalized_goal"], 
                            desired_goal = hindsight_goal, 
                            obs = f_trans_hindsight_goal.obs, 
                            action = f_trans_hindsight_goal.action, 
                            parent_info = parent_info,
                            env_info = tr.env_info) 
                    f_trans_hindsight_goal.done = achieved
                    f_trans_hindsight_goal.reward = f_reward
                    self._add_to_flat_replay_buffer(f_trans_hindsight_goal)

            # log undiscounted ep return to tensorboard (includes contribution 
            # from testing transitions but not from hindsight transitions!)
            # tensorboard logging
            if self._tb_writer is not None:
                self._tb_writer.add_scalar(f"{self.name}/ep_return", self._ep_return, sess_info.total_step)

        self._episode_transitions.clear()
            
    def get_algo_info(self, env_obs, parent_info):
        # child_be_deterministic instructs the children to be deterministic whereas
        # is_deterministic implies that this node is supposed to be deterministic
        if parent_info is not None:
            is_deterministic = parent_info.algo_info["child_be_deterministic"]
            child_be_deterministic = is_deterministic
        else:
            is_deterministic = False
            child_be_deterministic = False

        # if child_be_deterministic is true, all children have to use deterministic policies
        # until this node or an active parent gets back control (testing transition)
        child_be_deterministic = child_be_deterministic or np.random.rand() < self._testing_fraction
        new_algo_info = {
                "is_deterministic": is_deterministic, 
                "child_be_deterministic": child_be_deterministic
                }
        return new_algo_info



import numpy as np
from gym import spaces as gym_spaces

from . import Space

class DictSpace(Space):

    def __init__(self, space_dict):
        super().__init__()

        self._space_dict = space_dict

        # calculate dimension of flat space associated to this space
        self._flat_dim = self._get_flat_dim_uncached()

    def contains(self, value):
        if not isinstance(value, dict):
            return False
        for key, c in value.items():
            if not key in self._space_dict or not self._space_dict[key].contains(c):
                return False
        return True

    @property
    def length(self):
        return len(self._space_dict)

    def keys(self):
        return self._space_dict.keys()

    def get_flat_space(self, keys = None):
        if keys is None:
            keys = self._space_dict.keys()
        # NOTE: Need sorting in order to avoid different orders of keys.
        keys = sorted(keys)

        spaces = [self._space_dict[key].get_flat_space() for key in keys]

        return spaces[0].concatenate(*spaces)

    def _get_flat_dim_uncached(self, keys = None):
        if keys is None:
            keys = self._space_dict.keys()
        keys = sorted(keys)

        subspace_dims = [self._space_dict[key].get_flat_dim() for key in keys]

        return np.sum(subspace_dims)

    def get_flat_dim(self, keys = None):
        """Get dimension of flat space associated to this space."""
        # used cached value int this case. NOTE: This assumes the space is not modified after creation!
        if keys is None:
            return self._flat_dim
        else:
            return self._get_flat_dim_uncached(keys)

    def flatten_value(self, value, keys = None):
        if keys is None:
            keys = self._space_dict.keys()
        keys = sorted(keys)

        flat_sub_values = [self._space_dict[key].flatten_value(value[key]) for key in keys]
        flat_value = np.concatenate(flat_sub_values)
        return flat_value

    def unflatten_value(self, flat_value, keys = None):
        """Maps a vector from the space returned by flatten_value back to this space.
        
        Assumes that flat_value contains only contributions from the subspaces with keys 
        listed in keys. Assumes that space (and subspaces) have not been modified since 
        creation."""
        if keys is None:
            keys = self._space_dict.keys()
        keys = sorted(keys)

        # dimensions of subspaces
        dims = [self._space_dict[key].get_flat_dim() for key in keys]
        indices = [0]
        for dim in dims:
            indices.append(indices[-1] + dim)
        value = {key: self._space_dict[key].unflatten_value(flat_value[indices[i]:indices[i + 1]]) for i, key in enumerate(keys)}
        return value
        
    def sample(self):
        return {key: space.sample() for key, space in self._space_dict.items()}

    def get_gym_space(self):
        dict_of_gym_spaces = {key: space.get_gym_space() for key, space in self._space_dict.items()}
        return gym_spaces.Dict(dict_of_gym_spaces)

    def __eq__(self, other):
        if not (isinstance(other, DictSpace) and 
                self.keys() == other.keys()):
            return False
        for c1, c2 in zip(self._space_dict.values(), other._space_dict.values()):
            if c1 != c1:
                return False
        return True

    @staticmethod
    def concatenate(*args):
        space_dict = args[0]._space_dict
        for arg in args[1:]:
            space_dict.update(arg._space_dict)
        return DictSpace(component_spaces)

    @staticmethod
    def get_from_gym_space(gym_space):
        from .utils import space_from_gym_space
        assert isinstance(gym_space, gym_spaces.Dict)
        return DictSpace({key: space_from_gym_space(gs) for key, gs in gym_space.spaces.items()})

    def __str__(self):
        subspaces_string = ""
        for key, s in self._space_dict.items():
            if subspaces_string != "":
                subspaces_string += ", "
            subspaces_string += "{}: {}".format(key, s)
        return "DictSpace({{ {} }})".format(subspaces_string)

    def __getitem__(self, key):
        return self._space_dict[key]

    def get_subspace(self, keys):
        return DictSpace({key: space for key, space in self._space_dict.items() if key in keys})




import torch
from torch import nn
import torch.nn.functional as F

class NegativeNetwork(nn.Module):
    """Restricts output to negative components."""
    def __init__(self, core_module, device = "cpu"):
        """
        Args: 
            core_module: Module that is used to process inputs before sending through -exp.
        """
        super().__init__()

        self.core_module = core_module
        self.device = device
        self.nonlin = nn.LogSigmoid()


    def forward(self, x):
        if not isinstance(x, torch.Tensor):
            x = torch.tensor(x, device = self.device, dtype = torch.float)
        return self.nonlin(self.core_module(x))



import numpy as np
from gym import spaces as gym_spaces

from .space import Space

class MultiBinarySpace(Space):
    def __init__(self, n, dtype = np.int8):
        super().__init__()
        self.n = n
        self.dtype = dtype

    def contains(self, value):
        return np.all(np.logical_or(value == 0, value == 1))

    def get_flat_space(self):
        return self

    def flatten_value(self, value):
        return value

    def sample(self):
        res = np.random.randint(0, 2, size = self.n, dtype = np.int8)
        return np.array(res, dtype = self.dtype)

    def get_gym_space(self):
        return gym_spaces.MultiBinary(n, dtype = self.dtype)

    def __eq__(self, other):
        if not isinstance(other, MultiBinarySpace):
            return False
        return self.n == self.other and self.dtype == other.dtype

    @staticmethod
    def concatenate(*args):
        for arg in args[1:]:
            assert arg.dtype == args[0].dtype
        return MultiBinarySpace(sum([arg.n for arg in args]), dtype = args[0].dtype)

    @staticmethod
    def get_from_gym_space(gym_space):
        return MultiBinarySpace(gym_space.n, dtype = gym_space.dtype)

    def __str__(self):
        return "MultiBinarySpace: n = {}; dtype = {}".format(self.n, self.dtype)







import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    name="Graph_RL",
    version="0.1.1",
    description="Hierarchical reinforcement learning framework which uses a directed graph to define the hierarchy.",
    author="Nico Gürtler",
    author_email="nico.guertler@tuebingen.mpg.de",
    long_description=README,
    long_description_content_type="text/markdown",
    license="MIT",
    url="https://github.com/nicoguertler/graph_rl",
    keywords=["reinforcement learning", "hierarchical reinforcement learning"],
    packages=find_packages(),
    install_requires=["numpy", "gym", "tianshou==0.3.1", "pyglet", "dyn_rl_benchmarks"]
)




from abc import ABC, abstractmethod

class AuxReward:

    def __init__(self):
        pass

    @abstractmethod
    def __call__(self, obs, action, env_reward):
        pass

    def update(self, env_info):
        """This is called in every time step."""

        pass



import numpy as np

from . import AuxReward

class PowerReward(AuxReward):
    """Auxiliary reward proportional to negative power output."""

    def __init__(self, factor, map_to_vel, map_to_force):
        self._c = factor
        self._map_to_vel = map_to_vel
        self._map_to_force = map_to_force

    def __call__(self, obs, action, env_reward):
        vel = self._map_to_vel(obs)
        force = self._map_to_force(obs, action)
        return -self._c*np.dot(vel, force)



import torch
from torch import nn
import torch.nn.functional as F

class ConcatNetwork(nn.Module):
    """Concatenates all inputs."""
    def __init__(self, core_module, device = "cpu"):
        """
        Args: 
            core_module: Module that is used to process inputs once they have been concatenated.
        """
        super(ConcatNetwork, self).__init__()

        self.core_module = core_module
        self.device = device


    def forward(self, *args):
        inputs = []
        for arg in args:
            if not isinstance(arg, torch.Tensor):
                inputs.append(torch.tensor(arg, device = self.device, dtype = torch.float))
            else:
                inputs.append(arg)
        return self.core_module(torch.cat(inputs, dim = 1))



from .space import Space
from .gym_conv_dict import gym_conv_dict

def space_from_gym_space(gym_space):

    if not isinstance(gym_space, Space):
        cls = gym_space.__class__
        return gym_conv_dict[cls].get_from_gym_space(gym_space)
    else:
        return gym_space




import numpy as np

from . import Space

class TupleSpace(Space):

    def __init__(self, *args):
        super().__init__()

        self._component_spaces = args

    def contains(self, value):
        for c, c_space in zip(value, self._component_spaces):
            if not c_space.contains(c):
                return False
        return True

    @property
    def length(self):
        return len(self._component_spaces)

    def get_flat_space(self, indices = None):
        if indices is None:
            indices = range(self.length)

        spaces = [self._component_spaces[i] for i in indices]

        for space in spaces[1:]:
            assert isinstance(space, type(spaces[0]))

        return spaces[0].concatenate(*spaces)

    def flatten_value(self, value):
        flat_value = np.concatenate(
                [space.flatten_value(c) for c, space in zip(value, self._component_spaces)])
        return flat_value

    def sample(self):
        return tuple(space.sample() for space in self._component_spaces)

    def get_gym_space(self):
        raise NotImplementedError

    def __eq__(self, other):
        if not (isinstance(other, TupleSpace) and 
                self.length == other.length):
            return False
        for c1, c2 in zip(self._component_spaces, other._component_spaces):
            if c1 != c1:
                return False
        return True

    @staticmethod
    def concatenate(*args):
        component_spaces = []
        for arg in args:
            component_spaces.extend(arg._component_spaces)
        return TupleSpace(component_spaces)

    @staticmethod
    def get_from_gym_space(gym_space):
        raise NotImplementedError

    def __str__(self):
        res = "TupleSpace("
        for c in self._component_spaces:
            res += str(c)
        res += ")"
        return res
        
